// file      : xsd/cxx/parser/attribute-validation-template.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/attribute-validation-template.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      struct Test: Traversal::Attribute,
                   Traversal::AnyAttribute,
                   protected virtual Context
      {
        Test (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (SemanticGraph::Attribute& a)
        {
          String const& name (a.name ());

          if (a.qualified ())
          {
            String const& ns (a.namespace_ ().name ());

            os << "n == " << L << "\"" << name << "\" &&" << endl
               << "ns == " << L << "\"" << ns << "\"";
          }
          else
            os << "n == " << L << "\"" << name << "\" && ns.empty ()";
        }

        virtual Void
        traverse (SemanticGraph::AnyAttribute& a)
        {
          String ns (a.definition_namespace ().name ());

          for (SemanticGraph::Any::NamespaceIterator i (a.namespace_begin ()),
                 e (a.namespace_end ()); i != e;)
          {
            if (*i == L"##any")
            {
              os << "!n.empty ()";
            }
            else if (*i == L"##other")
            {
              if (ns)
              {
                // Note that here I assume that ##other does not include
                // unqualified names in a schema with target namespace.
                // This is not what the spec says but that seems to be
                // the consensus.
                //
                os << "(!ns.empty () && ns != " << L << "\"" << ns << "\")";
              }
              else
                os << "!ns.empty ()";
            }
            else if (*i == L"##local")
            {
              os << "(ns.empty () && !n.empty ())";
            }
            else if (*i == L"##targetNamespace")
            {
              os << "ns == " << L << "\"" << ns << "\"";
            }
            else
            {
              os << "ns == " << L << "\"" << *i << "\"";
            }

            if (++i != e)
              os << " ||" << endl;
          }
        }
      };

      //
      //
      struct PhaseOne : Traversal::Attribute,
                        protected virtual Context
      {
        PhaseOne (Context& c)
            : Context (c), test_ (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));
          String inst (L"_xsd_" + name + L"__");

          os << "if (";

          test_.traverse (a);

          os << ")"
             << "{"
             << "if (this->" << inst << " && !v)"
             << "{"
             << "this->" << inst << "->pre ();"
             << "this->" << inst << "->_pre_impl ();"
             << "this->" << inst << "->_characters (s);"
             << "this->" << inst << "->_post_impl ();"
             << "this->_xsd_call_glue_" << name << " (this->" << inst << ");"
             << "}";

          if (!a.optional ())
            os << "this->_xsd_" << ename (a) << "_v_state_ = true;";

          os << "break;"
             << "}";
        }

      private:
        Test test_;
      };


      //
      //
      struct PhaseTwo : Traversal::AnyAttribute,
                        protected virtual Context
      {
        PhaseTwo (Context& c)
            : Context (c), test_ (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          os << "if (";

          test_.traverse (a);

          os << ")" << endl
             << "{"
             << "this->_any_attribute (ns, n, s);"
             << "return true;"
             << "}";
        }

      private:
        Test test_;
      };


      //
      //
      struct AttributeStateInit: Traversal::Attribute,
                                 protected virtual Context
      {
        AttributeStateInit (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          if (!a.optional ())
          {
            os << "this->_xsd_" << ename (a) << "_v_state_ = false;";
          }
        }
      };


      //
      //
      struct AttributeStateCheck: Traversal::Attribute,
                                  protected virtual Context
      {
        AttributeStateCheck (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          if (!a.optional ())
          {
            String ns (a.qualified () ? a.namespace_ ().name () : String ());

            os << "if (!this->_xsd_" << ename (a) << "_v_state_)" << endl
               << "this->_expected_attribute (" << endl
               << L << "\"" << ns << "\", " <<
              L << "\"" << a.name () << "\");"
               << endl;
          }
        }
      };

      //
      //
      struct Complex : Traversal::Complex,
                       protected virtual Context
      {
        Complex (Context& c)
            : Context (c),
              phase_one_ (c),
              phase_two_ (c),
              attribute_state_init_ (c),
              attribute_state_check_ (c)
        {
          names_phase_one_ >> phase_one_;
          names_phase_two_ >> phase_two_;

          names_attribute_state_init_ >> attribute_state_init_;
          names_attribute_state_check_ >> attribute_state_check_;
        }

        virtual Void
        traverse (Type& c)
        {
          Boolean has_att (has<Traversal::Attribute> (c));
          Boolean has_any (has<Traversal::AnyAttribute> (c));

          if (!has_att && !has_any)
            return;

          // @@ At the moment we are unable to generate proper code for
          // complex types with only wildcards (no templete args).
          {
            Boolean htp (false);
            HasTemplateParam t (*this, htp);
            t.traverse (c);

            if (!htp)
              return;
          }

          String const& name (ename (c));
          Boolean restriction (restriction_p (c));

          os <<"// Attribute validation functions for " << name << "." << endl
             <<"//" << endl
             << "namespace _xsd_" << name
             << "{";

          if (has_att)
          {
            // _attribute_impl_phase_one
            //
            os << "template < ";

            {
              TemplateParamDeclList template_param_decl_list (*this);
              template_param_decl_list.traverse (c);
            }

            os << " >" << endl
               << "bool " << name;

            {
              TemplateParamUseList template_param_use_list (*this);
              template_param_use_list.traverse (c);
            }

            os << "::" << endl
               << "_attribute_impl_phase_one (const " << string_type <<
              "& ns," << endl
               << "const " << string_type << "& n," << endl
               << "const " << string_type << "& s," << endl
               << "bool v)"
               << "{"
               << "while (true)"
               << "{";

            names (c, names_phase_one_);

            // Nothing matched - call our base or return false if there is
            // no base.
            //
            os << "if (!v)" << endl
               << "return ";

            if (c.inherits_p ())
            {
              {
                BaseType base_type (*this, name, false);
                Traversal::Inherits inherits_type (base_type);
                inherits (c, inherits_type);
              }

              os << "::_attribute_impl_phase_one (ns, n, s, v);";
            }
            else
              os << "false;";


            os << endl
               << "break;"
               << "}"; // while

            if (restriction)
            {
              {
                BaseType base_type (*this, name, false);
                Traversal::Inherits inherits_type (base_type);
                inherits (c, inherits_type);
              }

              os << "::_attribute_impl_phase_one (ns, n, s, true);";
            }

            os << "return true;"
               << "}";
          }


          if (has_any)
          {
            // _attribute_impl_phase_two
            //
            os << "template < ";

            {
              TemplateParamDeclList template_param_decl_list (*this);
              template_param_decl_list.traverse (c);
            }

            os << " >" << endl
               << "bool " << name;

            {
              TemplateParamUseList template_param_use_list (*this);
              template_param_use_list.traverse (c);
            }

            os << "::" << endl
               << "_attribute_impl_phase_two (const " << string_type <<
              "& ns," << endl
               << "const " << string_type << "& n," << endl
               << "const " << string_type << "& s)"
               << "{";

            names (c, names_phase_two_);

            // Nothing matched - call our base or return false if there is
            // no base.
            //
            os << "return ";

            if (c.inherits_p ())
            {
              {
                BaseType base_type (*this, name, false);
                Traversal::Inherits inherits_type (base_type);
                inherits (c, inherits_type);
              }

              os << "::_attribute_impl_phase_two (ns, n, s);";
            }
            else
              os << "false;";

            os << "}";
          }

          if (has_att)
          {
            //@@ OPT: I don't need to generate pre/post _validate functions
            // if I don't have any required attributes.
            //

            // _pre_a_validate
            //
            os << "template < ";

            {
              TemplateParamDeclList template_param_decl_list (*this);
              template_param_decl_list.dispatch (c);
            }

            os << " >" << endl
               << "void " << name;

            {
              TemplateParamUseList template_param_use_list (*this);
              template_param_use_list.dispatch (c);
            }

            os << "::" << endl
               << "_pre_a_validate ()"
               << "{";

            names (c, names_attribute_state_init_);

            // Call our base last.
            //
            if (c.inherits_p ())
            {
              {
                BaseType base_type (*this, name, false);
                Traversal::Inherits inherits_type (base_type);

                inherits (c, inherits_type);
              }

              os << "::_pre_a_validate ();";
            }

            os << "}";


            // _post_a_validate
            //
            os << "template < ";

            {
              TemplateParamDeclList template_param_decl_list (*this);
              template_param_decl_list.dispatch (c);
            }

            os << " >" << endl
               << "void " << name;

            {
              TemplateParamUseList template_param_use_list (*this);
              template_param_use_list.dispatch (c);
            }

            os << "::" << endl
               << "_post_a_validate ()"
               << "{";

            // Call our base first.
            //
            if (c.inherits_p ())
            {
              {
                BaseType base_type (*this, name, false);
                Traversal::Inherits inherits_type (base_type);

                inherits (c, inherits_type);
              }

              os << "::_post_a_validate ();";
            }

            names (c, names_attribute_state_check_);

            os << "}";
          }

          os << "}";
        }

      private:
        PhaseOne phase_one_;
        Traversal::Names names_phase_one_;

        PhaseTwo phase_two_;
        Traversal::Names names_phase_two_;

        AttributeStateInit attribute_state_init_;
        Traversal::Names names_attribute_state_init_;

        AttributeStateCheck attribute_state_check_;
        Traversal::Names names_attribute_state_check_;
      };
    }

    Void
    generate_attribute_validation_template (Context& ctx)
    {
      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names schema_names;

      Namespace ns (ctx);
      Traversal::Names names;

      schema >> sources >> schema;
      schema >> schema_names >> ns >> names;

      Complex complex (ctx);

      names >> complex;

      schema.dispatch (ctx.schema_root);
    }
  }
}

