// file      : xsd/cxx/tree/elements.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <boost/regex.hpp>

#include <backend-elements/regex.hxx>

#include <cxx/tree/elements.hxx>

namespace CXX
{
  namespace Tree
  {
    //
    //
    String cxx_uq_id_expr (L"/^[a-zA-Z_]\\w*$//");

    // Context
    //
    Context::
    Context (std::wostream& o,
             SemanticGraph::Schema& root,
             CLI::Options const& ops,
             Counts const& counts_)
        : CXX::Context (o,
                        root,
                        ops.value<CLI::char_type> (),
                        ops.value<CLI::include_with_brackets> (),
                        ops.value<CLI::include_prefix> (),
                        ops.value<CLI::export_symbol> (),
                        ops.value<CLI::namespace_map> (),
                        ops.value<CLI::namespace_regex> (),
                        ops.value<CLI::namespace_regex_trace> (),
                        ops.value<CLI::generate_inline> ()),
          options (options_),
          counts (counts_),
          any_type (any_type_),
          any_simple_type (any_simple_type_),
          flags_type (flags_type_),
          parser_type (parser_type_),
          xml_element_type (xml_element_type_),
          xml_attribute_type (xml_attribute_type_),
          regex_custom_type_map (regex_custom_type_map_),
          direct_custom_type_map (direct_custom_type_map_),
          options_ (ops),
          parser_type_ (L"::xsd::cxx::xml::dom::parser< " + char_type + L" >"),
          xml_element_type_ (L"::xsd::cxx::xml::dom::element< " + char_type + L" >"),
          xml_attribute_type_ (L"::xsd::cxx::xml::dom::attribute< " + char_type + L" >"),
          ns_scope (ns_scope_),
          scope (scope_),
          inner_name (inner_name_),
          ns_scope_stack (ns_scope_stack_),
          scope_stack (scope_stack_)
    {
      any_type = xs_ns_name () + L"::type";
      any_simple_type = xs_ns_name () + L"::simple_type";
      flags_type = xs_ns_name () + L"::flags";

      // Custom type mapping.
      //
      using BackendElements::Regex::perl_s;

      typedef Containers::Vector<NarrowString> Vector;

      // Direct custom type mapping.
      //
      {
        Vector const& v (ops.value<CLI::custom_type> ());

        for (Vector::ConstIterator i (v.begin ()), e (v.end ()); i != e; ++i)
        {
          String s (*i);

          if (s.empty ())
            throw InvalidCustomTypeMapping (s, "mapping string is empty");

          // Split the string in two parts at the last '='.
          //
          Size pos (s.rfind ('='));

          // If no delimiter found then both type and base are empty.
          //
          if (pos == String::npos)
          {
            direct_custom_type_map[s].type.clear ();
            direct_custom_type_map[s].base.clear ();
            continue;
          }

          String name (s, 0, pos);
          String rest (s, pos + 1);

          // See if we've got the base part after '/'.
          //
          pos = rest.rfind ('/');

          String type, base;

          if (pos != String::npos)
          {
            type.assign (rest, 0, pos);
            base.assign (rest, pos + 1, String::npos);
          }
          else
            type = rest;

          // type can be a potentially-qualified template-id. base is
          // an unqualified C++ name.
          //

          if (!base.empty () && !perl_s (base, cxx_uq_id_expr).empty ())
            throw InvalidCustomTypeMapping (s, "invalid C++ identifier");

          direct_custom_type_map[name].type = type;
          direct_custom_type_map[name].base = base;
        }
      }

      // Regex custom type mapping.
      //
      {
        Vector const& v (ops.value<CLI::custom_type_regex> ());

        for (Vector::ConstIterator i (v.begin ()), e (v.end ()); i != e; ++i)
        {
          String s (*i);

          if (s.empty ())
            throw InvalidCustomTypeMapping (s, "mapping string is empty");

          WideChar delimiter (s[0]);

          // First get pattern.
          //
          Size pos (s.find (delimiter, 1));

          if (pos == String::npos)
            throw InvalidCustomTypeMapping (s, "invalid format");

          String pat (s, 1, pos - 1);
          String rest (s, pos + 1);

          String type, base;

          // See if we've got type and base.
          //
          if (!rest.empty ())
          {
            pos = rest.find (delimiter);

            if (pos == String::npos)
              throw InvalidCustomTypeMapping (s, "invalid format");

            type.assign (rest, 0, pos);
            rest = String (rest, pos + 1);

            if (!rest.empty ())
            {
              pos = rest.find (delimiter);

              if (pos == String::npos)
                throw InvalidCustomTypeMapping (s, "invalid format");

              base.assign (rest, 0, pos);
              rest = String (rest, pos + 1);

              if (!rest.empty ())
                throw InvalidCustomTypeMapping (s, "invalid format");
            }
          }

          regex_custom_type_map.push_back (
            RegexCustomTypeMapInfo (pat, type, base));
        }
      }
    }

    Context::
    Context (Context& c)
        : CXX::Context (c),
          options (c.options),
          counts (c.counts),
          any_type (c.any_type),
          any_simple_type (c.any_simple_type),
          flags_type (c.flags_type),
          parser_type (c.parser_type),
          xml_element_type (c.xml_element_type),
          xml_attribute_type (c.xml_attribute_type),
          regex_custom_type_map (c.regex_custom_type_map),
          direct_custom_type_map (c.direct_custom_type_map),
          ns_scope (c.ns_scope),
          scope (c.scope),
          inner_name (c.inner_name),
          ns_scope_stack (c.ns_scope_stack),
          scope_stack (c.scope_stack)
    {
    }

    Context::
    Context (Context& c, std::wostream& o)
        : CXX::Context (c, o),
          options (c.options),
          counts (c.counts),
          any_type (c.any_type),
          any_simple_type (c.any_simple_type),
          flags_type (c.flags_type),
          parser_type (c.parser_type),
          xml_element_type (c.xml_element_type),
          xml_attribute_type (c.xml_attribute_type),
          regex_custom_type_map (c.regex_custom_type_map),
          direct_custom_type_map (c.direct_custom_type_map),
          ns_scope (c.ns_scope),
          scope (c.scope),
          inner_name (c.inner_name),
          ns_scope_stack (c.ns_scope_stack),
          scope_stack (c.scope_stack)
    {
    }

    Boolean Context::
    custom_type (SemanticGraph::Type const& t, String& r) const
    {
      if (!t.named ())
        return false;

      String const& name (t.name ());

      // First search the direct mapping.
      //
      {
        DirectCustomTypeMap::ConstIterator i (
          direct_custom_type_map.find (name));

        if (i != direct_custom_type_map.end ())
        {
          r = i->second.type;
          return true;
        }
      }


      // Second search the regex mapping.
      //
      for (RegexCustomTypeMap::ConstIterator
             i (regex_custom_type_map.begin ()),
             e (regex_custom_type_map.end ());
           i != e; ++i)
      {
        //@@ Should probably store precompiled regexs in the map instead of
        //   of doing it every time.
        //
        boost::basic_regex<WideChar> expr (i->pat);

        if (regex_match (name,
                         expr,
                         boost::match_default | boost::format_all))
        {
          // Empty type sub tells us to use the original name.
          //
          if (i->type_sub.empty ())
          {
            r.clear ();
            return true;
          }

          r = regex_merge (name,
                           expr,
                           i->type_sub,
                           boost::match_default | boost::format_all);
          return true;
        }
      }

      return false;
    }

    String Context::
    custom_type (SemanticGraph::Type const& t) const
    {
      String r;
      if (custom_type (t, r))
      {
        // Empty type name tells us to use the original name.
        //
        if (r.empty ())
          r = ename (t);
      }

      return r;
    }

    Boolean Context::
    renamed_type (SemanticGraph::Type const& t, String& r) const
    {
      if (!t.named ())
        return false;

      String const& name (t.name ());

      // First search the direct mapping.
      //
      {
        DirectCustomTypeMap::ConstIterator i (
          direct_custom_type_map.find (name));

        if (i != direct_custom_type_map.end ())
        {
          r = i->second.base;
          return true;
        }
      }


      // Second search the regex mapping.
      //
      for (RegexCustomTypeMap::ConstIterator
             i (regex_custom_type_map.begin ()),
             e (regex_custom_type_map.end ());
           i != e; ++i)
      {
        //@@ Should probably store precompiled regexs in the map instead of
        //   of doing it every time.
        //
        boost::basic_regex<WideChar> expr (i->pat);

        if (regex_match (name,
                         expr,
                         boost::match_default | boost::format_all))
        {
          if (!i->base_sub.empty ())
          {
            r = regex_merge (name,
                             expr,
                             i->base_sub,
                             boost::match_default | boost::format_all);
          }
          else
            r.clear ();

          return true;
        }
      }

      return false;
    }

    Void Context::
    update_ns_scope ()
    {
      ns_scope.clear ();

      Boolean first (true);

      for (NamespaceStack::Iterator i (ns_scope_stack.begin ());
           i != ns_scope_stack.end ();
           ++i)
      {
        // We only qualify names until the namespace level.
        //
        if (first)
          first = false;
        else
          ns_scope += L"::";

        ns_scope += *i;
      }
    }

    Void Context::
    update_scope ()
    {
      scope.clear ();

      Boolean first (true);

      for (ScopeStack::Iterator i (scope_stack.begin ());
           i != scope_stack.end ();
           ++i)
      {
        // We only qualify names until the namespace level.
        //
        if (first)
          first = false;
        else
          scope += L"::";

        scope += *i;
      }

      inner_name.clear ();

      if (!scope_stack.empty ())
        inner_name = scope_stack.back ();
    }

    // GenerateDefautCtor
    //
    GenerateDefaultCtor::
    GenerateDefaultCtor (Context& c, Boolean& generate)
        : Context (c), generate_ (generate)
    {
      *this >> inherits_ >> *this;
      *this >> names_ >> *this;
    }

    Void GenerateDefaultCtor::
    traverse (SemanticGraph::Type& t)
    {
      generate_ = true;
    }

    Void GenerateDefaultCtor::
    traverse (SemanticGraph::Enumeration& e)
    {
      generate_ = true;
    }

    Void GenerateDefaultCtor::
    traverse (SemanticGraph::Member& m)
    {
      if (!skip (m) && min (m) == 1 && max (m) == 1)
        generate_ = true;
    }


    // GenerateFromBaseCtor
    //
    GenerateFromBaseCtor::
    GenerateFromBaseCtor (Context& c, Boolean& generate)
        : traverser_ (c, generate)
    {
      inherits_ >> traverser_;
    }

    Void GenerateFromBaseCtor::
    traverse (SemanticGraph::Complex& c)
    {
      inherits (c, inherits_);
    }

    GenerateFromBaseCtor::Traverser::
    Traverser (Context& c, Boolean& generate)
        : Context (c), generate_ (generate)
    {
      *this >> inherits_ >> *this;
      *this >> names_ >> *this;
    }

    Void GenerateFromBaseCtor::Traverser::
    traverse (SemanticGraph::Member& m)
    {
      if (!skip (m) && min (m) == 1 && max (m) == 1)
        generate_ = true;
    }


    // FromBaseCtorArg
    //
    FromBaseCtorArg::
    FromBaseCtorArg (Context& c, Boolean arg)
        : Context (c), arg_ (arg)
    {
    }

    Void FromBaseCtorArg::
    traverse (SemanticGraph::Member& m)
    {
      if (skip (m)) return;

      // Note that we are not going to include attributes with
      // default values here. Instead we are going to default-
      // initialize them.
      //
      if (min (m) == 1 && max (m) == 1)
      {
        String const& name (ename (m));

        os << "," << endl
           << "const " << iscope_name (name, "type") << "&";

        if (arg_)
          os << " _xsd_" << name;
      }
    }

    //
    //
    CtorArgs::
    CtorArgs (Context& c, String const& name, Boolean arg)
        : Context (c), name_ (name), arg_ (arg), first_ (true)
    {
      *this >> inherits_ >> *this;
      *this >> names_ >> *this;
    }

    Void CtorArgs::
    traverse (SemanticGraph::Type& t)
    {
      os << comma () << "const " << name_ << "::base_& " <<
        (arg_ ? name_ : String (""));
    }

    Void CtorArgs::
    traverse (SemanticGraph::Enumeration& e)
    {
      os << comma () << "const " << name_ << "::base_& " <<
        (arg_ ? name_ : String (""));
    }

    Void CtorArgs::
    traverse (SemanticGraph::Member& m)
    {
      if (skip (m)) return;

      // Note that we are not going to include attributes with
      // default values here. Instead we are going to default-
      // initialize them.
      //
      if (min (m) == 1 && max (m) == 1)
      {
        String const& name (ename (m));

        os << comma () << "const " << iscope_name (name, "type") << "&";

        if (arg_)
          os << " _xsd_" << name;
      }
    }

    String CtorArgs::
    comma ()
    {
      Boolean tmp (first_);
      first_ = false;
      return tmp ? "" : ",\n";
    }


    // GlobalElementBase
    //
    Boolean GlobalElementBase::
    generate_p (SemanticGraph::Element& e)
    {
      if (e.substitutes_p () &&
          ctx_.options.value<CLI::generate_polymorphic> ())
        return true;

      if (!doc_root_p (e))
        return false;

      // If we are not generating parsing/serialization code and this
      // element is of named type then we won't generate anything
      // from it.
      //
      if (ctx_.options.value<CLI::suppress_parsing> () &&
          !ctx_.options.value<CLI::generate_serialization> () &&
          e.type ().named ())
        return false;

      return true;
    }

    Boolean GlobalElementBase::
    doc_root_p (SemanticGraph::Element& e)
    {
      if (!ctx_.options.value<CLI::root_element_first> () &&
          !ctx_.options.value<CLI::root_element_last> () &&
          !ctx_.options.value<CLI::root_element_all> () &&
          !ctx_.options.value<CLI::root_element_none> () &&
          ctx_.options.value<CLI::root_element> ().empty ())
        return true; // By default treat them all.

      if (ctx_.options.value<CLI::root_element_none> ())
        return false;

      if (ctx_.options.value<CLI::root_element_all> ())
        return true;

      if (ctx_.options.value<CLI::root_element_first> () &&
          e.context ().count ("first") != 0)
        return true;

      if (ctx_.options.value<CLI::root_element_last> () &&
          e.context ().count ("last") != 0)
        return true;

      typedef Cult::Containers::Vector<NarrowString> Names;
      Names const& names (ctx_.options.value<CLI::root_element> ());

      // Hopefully nobody will specify more than a handful of names ;-).
      //
      for (Names::ConstIterator i (names.begin ()); i != names.end (); ++i)
      {
        String name (*i);

        if (e.name () == name)
          return true;
      }

      return false;
    }

    // Namespace
    //
    Namespace::
    Namespace (Context& c,
               Boolean set_scope,
               UnsignedLong first,
               UnsignedLong last)
        : CXX::Namespace (c, *this),
          GlobalElementBase (c),
          ctx_ (c),
          set_scope_ (set_scope),
          first_ (first),
          last_ (last),
          count_ (0)
    {
    }

    Void Namespace::
    traverse (Type& ns)
    {
      using SemanticGraph::Element;

      if (first_ > last_)
        CXX::Namespace::traverse (ns);
      else
      {
        Boolean opened (false);

        for (Type::NamesIterator i (ns.names_begin ());
             i != ns.names_end (); ++i)
        {
          SemanticGraph::Nameable& n (i->named ());

          if (n.is_a<SemanticGraph::Type> () ||
              (n.is_a<Element> () && generate_p (dynamic_cast<Element&> (n))))
          {
            if (count_ >= first_ && count_ <= last_)
            {
              if (!opened)
              {
                opened = true;
                pre (ns);
              }

              edge_traverser ().dispatch (*i);
            }

            ++count_;
          }
        }

        if (opened)
          post (ns);
      }
    }

    Void Namespace::
    enter (String const& name)
    {
      if (set_scope_)
        ctx_.enter_scope (name);

      ctx_.enter_ns_scope (name);
    }

    Void Namespace::
    leave ()
    {
      if (set_scope_)
        ctx_.leave_scope ();

      ctx_.leave_ns_scope ();
    }

    // Include
    //
    Void Includes::
    traverse (SemanticGraph::Path const& path)
    {
      using BackendElements::Regex::perl_s;

      // Note the use of the portable representation of the path.
      //
      String name (perl_s (String (path.string ()), expr_));

      Char op (include_with_brackets ? '<' : '"');
      Char cl (include_with_brackets ? '>' : '"');

      os << "#include " << op << include_prefix << name << cl << endl
         << endl;
    }
  }
}
