// file      : xsd/cxx/tree/serialization-header.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/serialization-header.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List: Traversal::List, protected virtual Context
      {
        List (Context& c,
              String const& name_ = "",
              String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
          {
            name = ename (l);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          // operator<< (xercesc::DOMElement)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xercesc::DOMElement&," << endl
             << "const " << type_name << "&);"
             << endl;

          // operator<< (xercesc::DOMAttr)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xercesc::DOMAttr&," << endl
             << "const " << type_name << "&);"
             << endl;

          // operator<< (list_stream)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xsd::cxx::tree::list_stream< " << char_type
             << " >&," << endl
             << "const " << type_name << "&);"
             << endl;

        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      struct Union: Traversal::Union, protected virtual Context
      {
        Union (Context& c,
               String const& name_ = "",
               String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
          {
            name = ename (u);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          // operator<< (xercesc::DOMElement)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xercesc::DOMElement&," << endl
             << "const " << type_name << "&);"
             << endl;

          // operator<< (xercesc::DOMAttr)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xercesc::DOMAttr&," << endl
             << "const " << type_name << "&);"
             << endl;

          // operator<< (list_stream)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xsd::cxx::tree::list_stream< " << char_type
             << " >&," << endl
             << "const " << type_name << "&);"
             << endl;
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      struct Enumeration: Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c,
                     String const& name_ = "",
                     String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
          {
            name = ename (e);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          // operator<< (xercesc::DOMElement)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xercesc::DOMElement&," << endl
             << type_name << ");"
             << endl;

          // operator<< (xercesc::DOMAttr)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xercesc::DOMAttr&," << endl
             << type_name << ");"
             << endl;

          // operator<< (list_stream)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xsd::cxx::tree::list_stream< " << char_type
             << " >&," << endl
             << type_name << ");"
             << endl;
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      //
      //
      template <typename Complex>
      struct Member : Traversal::Member, protected virtual Context
      {
        Member (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            // I need to build scope information with the same names
            // as in the header so that I get member escaping right.
            // Actual names will be from the public interface.
            //

            String const& name (ename (m));
            String type_name (scope
                              + L"::" + name
                              + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, name, type_name);
            Union union_ (*this, name, type_name);
            Complex complex (*this, name, type_name);
            Enumeration enumeration (*this, name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }

      private:
        String scope;
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c,
                 String const& name_ = "",
                 String const& type_name_ = L"")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
          {
            name = ename (c);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          // Go after anonymous types first.
          //
          {
            enter_scope (name);

            Member<Complex> member (*this, type_name);
            Traversal::Names names_member (member);

            names (c, names_member);

            leave_scope ();
          }


          // operator<< (xercesc::DOMElement)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xercesc::DOMElement&," << endl
             << "const " << type_name << "&);"
             << endl;

          Boolean simple (true);
          {
            IsSimpleType t (simple);
            t.dispatch (c);
          }

          if (simple)
          {
            // operator<< (xercesc::DOMAttr)
            //
            os << esymbol
               << "void" << endl
               << "operator<< (::xercesc::DOMAttr&," << endl
               << "const " << type_name << "&);"
               << endl;

            // operator<< (list_stream)
            //
            os << esymbol
               << "void" << endl
               << "operator<< (::xsd::cxx::tree::list_stream< " << char_type
               << " >&," << endl
               << "const " << type_name << "&);"
               << endl;
          }
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element,
                            GlobalElementBase,
                            protected virtual Context
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c), Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (!generate_p (m))
            return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named ())
          {
            String const& name (ename (m));
            String type_name (name + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, name, type_name);
            Union union_ (*this, name, type_name);
            Complex complex (*this, name, type_name);
            Enumeration enumeration (*this, name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };


      struct Serializer: Traversal::Element,
                         GlobalElementBase,
                         protected virtual Context
      {
        Serializer (Context& c)
            : GlobalElementBase (c),
              Context (c),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }


        virtual Void
        traverse (Type& e)
        {
          if (!doc_root_p (e))
            return;

          String const& name (ename (e));

          os << "// Serialize to an existing DOM instance." << endl
             << "//" << endl
             << esymbol
             << "void" << endl
             << name << " (::xercesc::DOMDocument&," << endl
             << "const " << type_name (e) << "&," << endl
             << flags_type << " = 0);"
             << endl
             << endl;

          os << "// Serialize to a new DOM instance." << endl
             << "//" << endl
             << esymbol
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument >" << endl
             << name << " (const " << type_name (e) << "&, " << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >&," << endl
             << flags_type << " = 0);"
             << endl
             << endl;

          os << "// Serialize to XMLFormatTarget." << endl
             << "//" << endl
             << esymbol
             << "void" << endl
             << name << " (::xercesc::XMLFormatTarget&," << endl
             << "const " << type_name (e) << "&, " << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >&," << endl
             << "const ::std::basic_string< " << char_type << " >& = " << L <<
            "\"UTF-8\"," << endl
             << flags_type << " = 0);"
             << endl
             << endl;

          os << esymbol
             << "void" << endl
             << name << " (::xercesc::XMLFormatTarget&," << endl
             << "const " << type_name (e) << "&, " << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >&," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >&," << endl
             << "const ::std::basic_string< " << char_type << " >& = " << L <<
            "\"UTF-8\"," << endl
             << flags_type << " = 0);"
             << endl;


          os << esymbol
             << "void" << endl
             << name << " (::xercesc::XMLFormatTarget&," << endl
             << "const " << type_name (e) << "&, " << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >&," << endl
             << "::xercesc::DOMErrorHandler&," << endl
             << "const ::std::basic_string< " << char_type << " >& = " << L <<
            "\"UTF-8\"," << endl
             << flags_type << " = 0);"
             << endl
             << endl;

          os << "// Serialize to std::ostream." << endl
             << "//" << endl
             << esymbol
             << "void" << endl
             << name << " (::std::ostream&," << endl
             << "const " << type_name (e) << "&, " << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >&," << endl
             << "const ::std::basic_string< " << char_type << " >& = " << L <<
            "\"UTF-8\"," << endl
             << flags_type << " = 0);"
             << endl
             << endl;

          os << esymbol
             << "void" << endl
             << name << " (::std::ostream&," << endl
             << "const " << type_name (e) << "&, " << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >&," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >&," << endl
             << "const ::std::basic_string< " << char_type << " >& = " << L <<
            "\"UTF-8\"," << endl
             << flags_type << " = 0);"
             << endl;

          os << esymbol
             << "void" << endl
             << name << " (::std::ostream&," << endl
             << "const " << type_name (e) << "&, " << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >&," << endl
             << "::xercesc::DOMErrorHandler&," << endl
             << "const ::std::basic_string< " << char_type << " >& = " << L <<
            "\"UTF-8\"," << endl
             << flags_type << " = 0);"
             << endl
             << endl;

        }

      private:
        String
        type_name (Type& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (t.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (t);

            return o.str ();
          }
          else
          {
            String const& name (ename (e));
            return name + L"::" + unclash (name, "type");
          }
        }

      private:
        String esymbol;
      };

    }

    Void
    generate_serialization_header (Context& ctx)
    {
      String c (ctx.char_type);

      ctx.os << "#include <iosfwd> // std::ostream&" << endl
             << endl
             << "#include <xercesc/dom/DOMDocument.hpp>" << endl
             << "#include <xercesc/dom/DOMErrorHandler.hpp>" << endl
             << "#include <xercesc/framework/XMLFormatter.hpp>" << endl
             << endl
             << "#include <xsd/cxx/xml/dom/auto-ptr.hxx>" << endl
             << "#include <xsd/cxx/tree/serialization.hxx>" << endl
             << endl;

      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names names_ns, names;

      Namespace ns (ctx, false);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);
      Serializer serializer (ctx);


      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;
      names >> serializer;

      schema.dispatch (ctx.schema_root);
    }
  }
}
