// file      : xsd/cxx/tree/tree-forward.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/tree-forward.hxx>
#include <cxx/tree/fundamental-header.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List,
                    protected virtual Context
      {
        List (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          String const& name (ename (l));

          if (String custom = custom_type (l))
          {
            String new_name;
            renamed_type (l, new_name);

            if (new_name)
              os << "class " << new_name << ";";

            if (custom == name)
              os << "class " << name << ";";
            else
              os << "typedef " << custom << " " << name << ";";
          }
          else
            os << "class " << name << ";";
        }
      };

      struct Union : Traversal::Union,
                     protected virtual Context
      {
        Union (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          String const& name (ename (u));

          if (String custom = custom_type (u))
          {
            String new_name;
            renamed_type (u, new_name);

            if (new_name)
              os << "class " << new_name << ";";

            if (custom == name)
              os << "class " << name << ";";
            else
              os << "typedef " << custom << " " << name << ";";
          }
          else
            os << "class " << name << ";";
        }
      };

      struct Enumeration : Traversal::Enumeration,
                           protected virtual Context
      {
        Enumeration (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String const& name (ename (e));

          if (String custom = custom_type (e))
          {
            String new_name;
            renamed_type (e, new_name);

            if (new_name)
              os << "class " << new_name << ";";

            if (custom == name)
              os << "class " << name << ";";
            else
              os << "typedef " << custom << " " << name << ";";
          }
          else
            os << "class " << name << ";";
        }
      };

      struct Complex : Traversal::Complex,
                       protected virtual Context
      {
        Complex (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& c)
        {
          String const& name (ename (c));

          if (String custom = custom_type (c))
          {
            String new_name;
            renamed_type (c, new_name);

            if (new_name)
              os << "class " << new_name << ";";

            if (custom == name)
              os << "class " << name << ";";
            else
              os << "typedef " << custom << " " << name << ";";
          }
          else
            os << "class " << name << ";";
        }
      };
    }

    Void
    generate_forward (Context& ctx,
                      String const& hxx_expr,
                      String const& fwd_expr)
    {
      String xml_schema (ctx.options.value<CLI::extern_xml_schema> ());

      // Inlcude or Emit fundamental types.
      //
      if (xml_schema)
      {
        using BackendElements::Regex::perl_s;

        // Note the use of the portable representation of the path.
        //
        String name (perl_s (xml_schema, hxx_expr));

        Char op (ctx.include_with_brackets ? '<' : '"');
        Char cl (ctx.include_with_brackets ? '>' : '"');

        ctx.os << "#include " << op << ctx.include_prefix << name << cl <<
          endl
               << endl;
      }
      else
      {
        ctx.os << "#include <xsd/cxx/tree/exceptions.hxx>" << endl
               << "#include <xsd/cxx/tree/elements.hxx>" << endl
               << "#include <xsd/cxx/tree/types.hxx>" << endl;

        if (!ctx.options.value<CLI::suppress_parsing> () ||
            ctx.options.value<CLI::generate_serialization> ())
        {
          ctx.os << "#include <xsd/cxx/xml/error-handler.hxx>" << endl
                 << endl;
        }

        if (ctx.options.value<CLI::generate_serialization> ())
          ctx.os << "#include <xsd/cxx/xml/dom/namespace-infomap.hxx>" << endl
                 << endl;

        Traversal::Schema schema, xsd;
        Traversal::Implies implies;
        Traversal::Names names;
        FundamentalNamespace ns (ctx);

        schema >> implies >> xsd >> names >> ns;

        schema.dispatch (ctx.schema_root);
      }

      // First emit header includes.
      //
      if (ctx.options.value<CLI::generate_forward> ())
      {
        Traversal::Schema schema;
        Includes includes (ctx, fwd_expr);

        schema >> includes;

        schema.dispatch (ctx.schema_root);
      }


      ctx.os << "// Forward declarations." << endl
             << "//" << endl;

      Traversal::Schema schema;
      Traversal::Sources sources;
      Traversal::Names names_ns, names;
      Namespace ns (ctx, false);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;

      schema.dispatch (ctx.schema_root);

      ctx.os << endl;
    }
  }
}
