// file      : xsd/cxx/tree/validator.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/validator.hxx>

#include <cult/containers/set.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cxx/tree/elements.hxx>

#include <iostream>

using std::wcerr;

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      class ValidationContext: public Context
      {
      public:
        ValidationContext (SemanticGraph::Schema& root,
                           CLI::Options const& options,
                           Counts const& counts,
                           Boolean& valid_)
            : Context (std::wcerr, root, options, counts),
              valid (valid_),
              classify_warning_issued (classify_warning_issued_),
              classify_warning_issued_ (false),
              subst_group_warning_issued (subst_group_warning_issued_),
              subst_group_warning_issued_ (false)
        {
        }
      public:
        String
        xpath (SemanticGraph::Nameable& n)
        {
          if (n.is_a<SemanticGraph::Namespace> ())
            return L"<namespace-level>"; // There is a bug if you see this.

          if (n.named ())
          {
            SemanticGraph::Scope& scope (n.scope ());

            if (scope.is_a<SemanticGraph::Namespace> ())
              return n.name ();

            return xpath (scope) + L"/" + n.name ();
          }
          else
          {
            return L"(anonymous type for " +
              n.context ().get<String> ("instance-name") + L")";
          }
        }

      protected:
        ValidationContext (ValidationContext& c)
            :  Context (c),
               valid (c.valid),
               classify_warning_issued (c.classify_warning_issued),
               subst_group_warning_issued (c.subst_group_warning_issued)
        {
        }

      protected:
        Boolean& valid;

        Boolean& classify_warning_issued;
        Boolean classify_warning_issued_;

        Boolean& subst_group_warning_issued;
        Boolean subst_group_warning_issued_;
      };


      //
      //
      struct Any : Traversal::Any, protected virtual ValidationContext
      {
        Any (ValidationContext& c)
            : ValidationContext (c)
        {
        }

        struct Element: Traversal::Element,
                        protected virtual ValidationContext
        {
          Element (ValidationContext& c, SemanticGraph::Any& any)
              : ValidationContext (c),
                any_ (any),
                ns_ (any.definition_namespace ().name ())
          {
          }

          virtual Void
          traverse (SemanticGraph::Element& e)
          {
	    if (skip (e)) return;

            using SemanticGraph::Any;

            Boolean q (e.qualified ());
            String ns (q ? e.namespace_ ().name () : "");

            for (Any::NamespaceIterator i (any_.namespace_begin ());
                 i != any_.namespace_end (); ++i)
            {
              Boolean failed (false);

              if (*i == L"##any")
              {
                failed = true;
              }
              else if (*i == L"##other")
              {
                if (ns_)
                {
                  // Note that here I assume that ##other does not
                  // include names without target namespace. This
                  // is not what the spec says but that seems to be
                  // the consensus.
                  //
                  failed = q && ns != ns_;
                }
                else
                {
                  // No target namespace.
                  //
                  failed = q && ns != L"";
                }
              }
              else if (*i == L"##local")
              {
                failed = !q || ns == L"";
              }
              else if (*i == L"##targetNamespace")
              {
                failed = (q && ns_ == ns) || (!q && ns_ == L"");
              }
              else
              {
                failed = q && *i == ns;
              }

              if (failed)
              {
                Any& a (any_);

                os << a.file () << ":" << a.line () << ":" << a.column ()
                   << ": error: namespace '" << *i << "' allows for element "
                   << "'" << e.name () << "'" << endl;

                os << e.file () << ":" << e.line () << ":" << e.column ()
                   << ": error: element '" << e.name () << "' is defined "
                   << "here" << endl;

                os << a.file () << ":" << a.line () << ":" << a.column ()
                   << ": error: such wildcards are not supported" << endl;

                valid = false;
              }
            }
          }

        private:
          SemanticGraph::Any& any_;
          String ns_;
        };

        struct Complex: Traversal::Complex
        {
          Complex ()
              : up_ (true), down_ (true)
          {
          }

          virtual Void
          post (Type& c)
          {
            // Go down the inheritance hierarchy.
            //
            if (down_)
            {
              Boolean up = up_;
              up_ = false;

              if (c.inherits_p ())
                dispatch (c.inherits ().base ());

              up_ = up;
            }

            // Go up the inheritance hierarchy.
            //
            if (up_)
            {
              Boolean down = down_;
              down_ = false;

              for (Type::BegetsIterator i (c.begets_begin ());
                   i != c.begets_end (); ++i)
              {
                dispatch (i->derived ());
              }

              down_ = down;
            }
          }

        private:
          Boolean up_, down_;
        };

        virtual Void
        traverse (SemanticGraph::Any& a)
        {
          using SemanticGraph::Compositor;

          // Find our complex type.
          //
          Compositor* c (&a.contained_particle ().compositor ());

          while(!c->contained_compositor_p ())
            c = &c->contained_particle ().compositor ();

          SemanticGraph::Complex& type (
            dynamic_cast<SemanticGraph::Complex&> (
              c->contained_compositor ().container ()));

          Complex complex;
          Traversal::Names names;
          Element element (*this, a);

          complex >> names >> element;

          complex.dispatch (type);
        }
      };


      //
      //
      struct Traverser : Traversal::Schema,
                         Traversal::Complex,
                         Traversal::Type,
                         Traversal::Element,
                         GlobalElementBase,
                         protected virtual ValidationContext
      {
        Traverser (ValidationContext& c)
            : GlobalElementBase (c),
              ValidationContext (c),
              anonymous_ (c, *this),
              any_ (c)
        {
          *this >> sources_ >> *this;
          *this >> schema_names_ >> ns_ >> names_;

          names_ >> *this;
          names_ >> anonymous_;

          *this >> names_;

          // Any
          //
          *this >> contains_compositor_ >> compositor_ >> contains_particle_;
          contains_particle_ >> compositor_;
          contains_particle_ >> any_;
        }

        virtual Void
        traverse (SemanticGraph::Complex& c)
        {
          using SemanticGraph::Schema;

          traverse (static_cast<SemanticGraph::Type&> (c));

          if (c.inherits_p ())
          {
            SemanticGraph::Type& t (c.inherits ().base ());

            if (t.named () &&
                types_.find (
                  t.scope ().name () + L"#" + t.name ()) == types_.end ())
            {
              // Don't worry about types that are in included/imported
              // schemas.
              //
              Schema& s (dynamic_cast<Schema&> (t.scope ().scope ()));

              if (&s == &schema_root || sources_p (schema_root, s))
              {
                valid = false;

                os << c.file () << ":" << c.line () << ":" << c.column ()
                   << ": error: type '" << xpath (c) << "' inherits from "
                   << "yet undefined type '" << xpath (t) << "'" << endl;

                os << t.file () << ":" << t.line () << ":" << t.column ()
                   << ": info: '" << xpath (t) << "' is defined here"
                   << endl;

                os << c.file () << ":" << c.line () << ":" << c.column ()
                   << ": info: inheritance from a yet-undefined type is "
                   << "not supported" << endl;

                os << c.file () << ":" << c.line () << ":" << c.column ()
                   << ": info: re-arrange your schema and try again"
                   << endl;
              }
            }
          }

          Traversal::Complex::traverse (c);
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          using SemanticGraph::Member;
          using SemanticGraph::Type;

          // This is also used to traverse Complex.
          //

          if (t.named ())
          {
            types_.insert (t.scope ().name () + L"#" + t.name ());
          }
          else
          {
            // Check for excessive classification.
            //
            if (excessive_classify_.find (&t) == excessive_classify_.end ())
            {
              // We cannot use iterator difference because enumerators
              // belongs to their enumeration type.
              //
              UnsignedLong n (0);

              for (Type::ClassifiesIterator i (t.classifies_begin ());
                   i != t.classifies_end (); ++i)
              {
                if (Member* m = dynamic_cast<Member*> (&i->instance ()))
                {
                  if (count_member (*m))
                    ++n;
                }
              }

              if (n > 1)
              {
                if (!classify_warning_issued)
                {
                  classify_warning_issued = true;

                  os << t.file ()
                     << ": warning: detected anonymous types that are "
                     << "used to classify multiple attributes/elements"
                     << endl;

                  os << t.file ()
                     << ": warning: this may result in excessive code "
                     << "duplication" << endl;

                  if (!options.value<CLI::show_classification> ())
                    os << t.file ()
                       << ": info: use --show-classification option to "
                       << "see these types" << endl;

                  os << t.file ()
                     << ": info: consider explicitly naming these types "
                     << "or use --morph-anonymous option"
                     << endl;
                }

                if (options.value<CLI::show_classification> ())
                {
                  os << t.file () << ":" << t.line () << ":" << t.column ()
                     << ": warning: '" << xpath (t) << "' is used to "
                     << "classify " << n << " attributes/elements:" << endl;

                  for (Type::ClassifiesIterator i (t.classifies_begin ());
                       i != t.classifies_end (); ++i)
                  {
                    if (Member* m = dynamic_cast<Member*> (&i->instance ()))
                    {
                      if (count_member (*m))
                        os << m->file () << ":" << m->line () << ":"
                           << m->column () << ": warning: '" << m->name ()
                           << "', which is defined here" << endl;
                    }
                  }
                }

                excessive_classify_.insert (&t);
              }
            }
          }
        }

        //
        //
        Boolean
        count_member (SemanticGraph::Member& m)
        {
          using SemanticGraph::Schema;
          using SemanticGraph::Namespace;
          using SemanticGraph::Element;
          using SemanticGraph::Attribute;

          if (skip (m))
            return false;

          // Don't count namespace-scope members for which we
          // are not generating anything.
          //
          if (Namespace* ns = dynamic_cast<Namespace*> (&m.scope ()))
          {
            // Don't count global attributes.
            //
            if (m.is_a<Attribute> ())
              return false;

            Schema& s (dynamic_cast<Schema&> (ns->scope ()));

            // Don't count global elements/attributes from
            // imported/included schemas.
            //
            if (&s != &schema_root && !sources_p (schema_root, s))
              return false;

            if (!generate_p (dynamic_cast<Element&> (m)))
              return false;
          }

          return true;
        }

        // Return true if root sources s.
        //
        Boolean
        sources_p (SemanticGraph::Schema& root, SemanticGraph::Schema& s)
        {
          using SemanticGraph::Schema;
          using SemanticGraph::Sources;

          for (Schema::UsesIterator i (root.uses_begin ());
               i != root.uses_end (); ++i)
          {
            if (i->is_a<Sources> ())
            {
              if (&i->schema () == &s || sources_p (i->schema (), s))
                return true;
            }
          }

          return false;
        }

        virtual Void
        traverse (SemanticGraph::Element& e)
        {
          // Note that there is no test for generate_p since we want
          // to catch cases when things are not being generated but
          // most likely should have been.
          //
          if (e.substitutes_p () &&
              !options.value<CLI::generate_polymorphic> () &&
              !subst_group_warning_issued)
          {
            subst_group_warning_issued = true;

            os << e.file () << ":" << e.line () << ":" << e.column ()
               << ": warning: substitution groups are used but "
               << "--generate-polymorphic was not specified" << endl;

            os << e.file () << ":" << e.line () << ":" << e.column ()
               << ": info: generated code may not be able to handle "
               << "some conforming instances" << endl;
          }
        }

      private:
        Containers::Set<String> types_;
        Containers::Set<SemanticGraph::Type*> excessive_classify_;

        Traversal::Sources sources_;

        Traversal::Names schema_names_;
        Traversal::Namespace ns_;

        Traversal::Names names_;

        // Anonymous feedback via belongs edge.
        //
        struct Anonymous : AnonymousBase,
                           protected virtual ValidationContext
        {
          Anonymous (ValidationContext& c, Traversal::NodeDispatcherBase& d)
              : ValidationContext (c),
                AnonymousBase (d)
          {
          }

          virtual void
          member_pre (SemanticGraph::Member& m)
          {
            m.type ().context ().set ("instance-name", xpath (m));
          }

          virtual void
          member_post (SemanticGraph::Member& m)
          {
            m.type ().context ().remove ("instance-name");
          }

        } anonymous_;


        // Any.
        //
        Any any_;
        Traversal::Compositor compositor_;
        Traversal::ContainsParticle contains_particle_;
        Traversal::ContainsCompositor contains_compositor_;
      };
    }

    Validator::
    Validator ()
    {
      // Dummy ctor, help with long symbols on HP-UX.
    }

    Boolean Validator::
    validate (CLI::Options const& options,
              SemanticGraph::Schema& schema,
              SemanticGraph::Path const& tu,
              Counts const& counts)
    {
      //
      //
      if (options.value<CLI::char_type> () != "char" &&
          options.value<CLI::char_type> () != "wchar_t")
      {
        wcerr << "warning: unknown base character type '" <<
          options.value<CLI::char_type> ().c_str () << "'" << endl;
      }

      //
      //
      if (options.value<CLI::generate_polymorphic> () &&
          (options.value<CLI::generate_insertion> () ||
           options.value<CLI::generate_extraction> ()))
      {
        wcerr << "warning: stream insertion/extraction operations "
              << "are not yet polymorphism-aware" << endl;
      }

      // Issue a warning we there are more than one global element
      // and we are generating parsing/serialization functions for
      // all of them by default.
      //
      Boolean par (!options.value<CLI::suppress_parsing> ());
      Boolean ser (options.value<CLI::generate_serialization> ());

      if (counts.global_elements > 1 &&
          (par || ser) &&
          !options.value<CLI::root_element_first> () &&
          !options.value<CLI::root_element_last> () &&
          !options.value<CLI::root_element_all> () &&
          !options.value<CLI::root_element_none> () &&
          options.value<CLI::root_element> ().empty ())
      {
        wcerr << schema.file () << ": warning: generating " <<
          (par ? "parsing " : "") <<
          (ser ? (par ? "and serialization " : "serialization ") : "") <<
          "functions for " << counts.global_elements << " global elements"
              << endl;

        wcerr << schema.file () << ": info: use --root-element-* options "
              << "to specify document root(s)" << endl;
      }

      Boolean valid (true);

      ValidationContext ctx (schema, options, counts, valid);

      Traverser traverser (ctx);

      traverser.dispatch (schema);

      return valid;
    }
  }
}

